import numpy as np
from sslf.sslf import Spectrum
from scipy.signal import find_peaks
from scipy.optimize import curve_fit

def find_lines(flux):
    """ Create a line finder to find significant points """
    s = Spectrum(flux)
    s.find_cwt_peaks(scales=np.arange(4,10), snr=3)
    spec_peaks = s.channel_peaks

    # Calculate the ratio of the snrs and scales
    snrs = [round(i,2) for i in s.peak_snrs] # the snrs of the peaks
    scales = [i[1]-i[0] for i in s.channel_edges] # the scales of the peaks
    return spec_peaks, snrs, scales

def gaussf(x, a, s, x0):
    """ Function to fit a sum of gaussians """
    y = 0
    for i in range(1,30):
        y += (a * np.exp(-((x-i*x0) / s)**2)) # i = 1,2,3 ... 9, 10
    return y

def find_params(frequency, flux, loc):
    """ Find the best fitting parameters for the gaussians """
    params, covars = curve_fit(lambda x, a, s: gaussf(x, a, s, x0=loc), 
        frequency, flux, bounds=[[0, (1/8)], [2*max(flux), (2/3)]], absolute_sigma=True) # best fit
    return params, covars

def penalise_chi2(gauss_peaks, blind_lines):
    """ Penalise chi-squared values that do not fit to lines found with sslf line finder """

    # Variable Settings
    gauss_peaks = np.sort(gauss_peaks)
    blind_lines = np.sort(blind_lines)
    
    num_gaussians = len(gauss_peaks)
    num_blind_lines = len(blind_lines)
    penalising_factor = 1.2
    tolerance = 3

    # If the number of lines are not equal, return penalising factor
    if num_blind_lines != num_gaussians:
        return penalising_factor

    # If the lines are not within a range of each other, return penalising factor
    for bline, gline in zip(blind_lines, gauss_peaks):
        if gline not in range(bline-tolerance, bline+tolerance):
            return penalising_factor
    else:
        return 1 # No penalisation found, return factor of 1

def calc_reduced_chi2(flux, f_exp, gauss_peaks, uncertainty):
    """ Calculate the reduced chi-squared: chi2_r = chi2 / (num_points - free_params - 1) """

    # Calculate the degrees of freedom
    flux_points = len(flux)
    num_gauss = len(gauss_peaks)
    gauss_dofs = 2*num_gauss # 2 DoF per gaussian (amp, std)
    DoF = flux_points - gauss_dofs - 1

    # Calculate chi-squared
    chi2 = sum(((flux - f_exp) / uncertainty)**2)

    # Calculate reduced chi-squared
    reduced_chi2 = chi2 / DoF
    return reduced_chi2

class flux_zfind(object):
    def __init__(self, transition, frequency, flux, uncertainty=1):
        """
        Find statistics of the source via gaussian overlays
        
        Parameters
        ----------
        transition : float
            The first transition frequency of the element or molecule to search for. Units
            follow frequency
        
        frequency : list
            A list of frequency values calculated from fits2flux
        
        flux : list
            A list of flux values calculated from fits2flux

        uncertainty : list, optional
            A list of uncertainty values calculated from fits2flux. If left unset, resulting 
            chi-squared statistics will be relative. Default = 1
        """
        
        self.transition = transition
        self.frequency = frequency
        self.flux = flux
        self.uncertainty = uncertainty
    
    def gauss_zfind(self, z_start=0, dz=0.01, z_end=10, sigma=1, penalise=True):
        """ 
        Finds the best redshift by fitting gaussian functions overlayed on flux data. The
        chi-squared is calculated at every redshift by iterating through delta-z. The most 
        likely redshift corresponds to the minimum chi-squared.
        
        Parameters
        ----------
        z_start : int, optional
            The first value in the redshift list. Default = 0
            
        dz : float, optional
            The change in redshift. Default = 0.01
        
        z_end : int, optional
            The final value in the redshift list. Default = 10
        
        sigma : float, optional
            The significance level of the uncertainty in the redshift 
            found at the minimum chi-squared. Default = 1
            
        penalise : bool, optional
            If True, perform chi-squared penalisation with sslf. Default = True
        
        Returns
        -------
        z : list
            The list of redshifts that was used to calculate the chi-squared
        
        chi2 : list
            A list of calculated chi-squared values
        """
        
        # Initialise
        z = np.arange(z_start, z_end+dz, dz)
        sigma = sigma
        all_chi2 = []
        all_params = []
        all_num_peaks = []
        all_perrs = []
        self.blind_lines, self.snrs, self.scales = find_lines(self.flux)
        
        # Interate through the list of redshifts and calculate the chi-squared
        for dz in z:

            # Calculate the offset of the gaussians
            loc = self.transition/(1+dz)

            # Calculate parameters of gaussian fitS
            params, covars = find_params(self.frequency, self.flux, loc)
            perr = np.sqrt(np.diag(covars)) # calculate the error on the gaussian parameters
            
            # Calculate the expected flux array (gaussian overlay)
            f_exp = gaussf(self.frequency, a=params[0], s=params[1], x0=loc)

            # calculate the number of gaussians overlayed
            gauss_peaks = find_peaks(f_exp)[0]

            # Calculate the reduced chi-squared and check if it should be penalised
            if penalise:
                multiplier = penalise_chi2(gauss_peaks, self.blind_lines)
            else:
                multiplier = 1
            reduced_chi2 = calc_reduced_chi2(self.flux, f_exp, gauss_peaks, self.uncertainty) * multiplier

            all_chi2.append(reduced_chi2)
            all_params.append(params)
            all_num_peaks.append(len(gauss_peaks))
            all_perrs.append(perr)
        
        lowest_index = np.argmin(all_chi2)
        self.params = all_params[lowest_index]
        self.perr = all_perrs[lowest_index]
        self.peaks = all_num_peaks[lowest_index]
        return z, all_chi2
    
    def gauss_params(self):
        """
        Get the paramters of the best fitting redshift. Must run gauss_zfind first.
        
        Returns
        -------
        params : list
            Best fitting redshift parameters -> [amplitude, standard deviation]
        
        perr : list
            Errors on the best fitting redshift parameters -> [amplitude, standard deviation] 
        """
        return self.params, self.perr
    
    def gauss_sslf(self):
        """ 
        Get the statistics of the lines found with sslf. Must run gauss_zfind first.
        
        Returns
        -------
        peaks : list
            The channel locations of each significant points.
        
        snrs : list
            The signal-to-noise ratio of each significant points.
        
        scales : list
            The scale of each significant point
        """
        return self.blind_lines, self.snrs, self.scales   


import datetime
import numpy as np

from time import time
from random import random
from astropy.io import fits
from astropy.wcs import WCS
from flux_zfind import find_lines
from warnings import filterwarnings
from scipy.spatial.distance import cdist
from photutils.background import Background2D
from PyAstronomy.pyasl import hmsToDeg, dmsToDeg, degToHMS, degToDMS
from photutils.aperture import CircularAperture, CircularAnnulus, ApertureStats, aperture_photometry

def wcs2pix(ra, dec, hdr):
    """ Convert right ascension and declination to x, y positional world coordinates """

    w = WCS(hdr) # Get the world coordinate system

    # If there are more than 2 axis, drop them
    if hdr['NAXIS'] > 2:
        w = w.dropaxis(3) # stokes
        w = w.dropaxis(2) # frequency

    # Convert to decimal degrees
    ra = hmsToDeg(ra[0], ra[1], ra[2])
    dec = dmsToDeg(dec[0], dec[1], dec[2], esign=dec[3])

    # Convert world coordinates to pixel
    x, y = w.all_world2pix(ra, dec, 1)

    # Round to nearest integer
    x = int(np.round(x))
    y = int(np.round(y))
    return [x, y]

def pix2wcs(x, y, hdr):
    """ Convert x,y positional coordinates to world ra, dec"""
    # Drop redundant axis
    w = WCS(hdr)
    if hdr['NAXIS'] > 2:
        w = w.dropaxis(3) # stokes
        w = w.dropaxis(2) # frequency

    # Convert x,y to degrees
    ra, dec = w.all_pix2world(x, y, 1)

    # Convert degrees to ra, dec
    ra = degToHMS(ra)
    dec = degToDMS(dec)
    return ra, dec

def get_sci_exponent(number):
    """ Find the scientific exponent of a number """
    abs_num = np.abs(number)
    base = np.log10(abs_num) # Log rules to find exponent
    exponent = int(np.floor(base)) # convert to floor integer
    return exponent

def get_eng_exponent(number):
    """ 
    Find the nearest power of 3 (lower). In engineering format,
    exponents are multiples of 3.
    """
    
    # First get the scientific exponent
    exponent = get_sci_exponent(number)
    
    # If the exponent is a multiple of 3, return it
    for i in range(3):
        
        if exponent > 0:
            unit = exponent-i
        else:
            unit = exponent+i
            
        if unit % 3 == 0:
            return unit

def average_zeroes(array):
    """ Average zeroes with left & right values in a list """
    for i, val in enumerate(array):
        if val == 0:
            try:
                array[i] = (array[i-1] + array[i+1])/2
            except IndexError:
                array[i] = (array[i-2] + array[i-1])/2
    return array

def _calc_beam_area(bvalue):
    """ Caclulate the corrected beam area (from Jy/beam to Jy) """
    bmaj = bvalue/3600
    bmin = bmaj
    barea = 1.1331 * bmaj * bmin
    return barea

def spaced_circle_points(num_points, circle_radius, centre_coords, minimum_spread_distance):
    """ Calculate points in a circle with a minimum spread """
    
    points = [centre_coords] # centre_coords = [x,y] -> points = [[x,y]]
    
    # Iterate through the number of points.
    for _ in range(num_points-1):
        
        # Keep generating the current point until it is at least the minimum distance away from all 
        while True:
            theta = 2 * np.pi * random() # choose a random direction
            r = circle_radius * random() # choose a random radius

            # Convert coordinates to cartesian
            x = r * np.cos(theta) + centre_coords[0]
            y = r * np.sin(theta) + centre_coords[1]

            # Find the distance between all the placed points
            distances = cdist([[x,y]], points, 'euclidean')
            min_distance = min(distances[0])
            
            # If the minimum distance is satisfied for all points, go to next point
            if min_distance >= minimum_spread_distance or len(points) == 1:
                points.append([x,y])
                break
    return points

def calc_pixels(spec_peaks, flux):
    """ 
    Calculate the number of pixels above 0 Jy left & right of all lines found 
    by the sslf line finder.
    """
    pixels = []
    for peak in spec_peaks:
        num_pix = 0

        # left side
        i = 1
        try:
            while flux[peak-i] > 0:
                num_pix += 1
                i += 1
        except IndexError:
            pass

        # right side
        i = 0
        try:
            while flux[peak+i] > 0:
                num_pix += 1
                i += 1
        except IndexError:
            pass

        pixels.append(num_pix)
    return pixels

class fits2flux(object):
    def __init__(self, image, ra, dec, aperture_radius, bvalue):
        """ 
        Calculate the flux over the frequency range of a .fits image
        
        Parameters
        ----------
        image : .fits
            A .fits image file
        
        ra : list
           Right ascension of the target [h, m, s]
        
        dec : list
            Declination of the target [d, m, s, esign]
        
        aperture_radius : float
            Radius of the aperture to use over the source (pixels)

        bvalue : float
            The value of BMAJ and BMIN (arcseconds)
        """
        image = fits.open(image)
        self.hdr = image[0].header
        self.data = image[0].data[0]
        self.ra = ra
        self.dec = dec
        self.aperture_radius = aperture_radius
        self.bvalue = bvalue

        # Ignore warnings (there are hundreds)
        filterwarnings("ignore", module='photutils.background')
        filterwarnings("ignore", module='astropy.wcs.wcs')

    def get_freq(self):
        """ 
        Caclulate the frequency axis (x-axis) of the flux

        Returns
        -------
        frequency : list
            A list of frequencies corresponding to individual channels of a .fits image
        """
        
        # Required values
        start = self.hdr['CRVAL3']
        increment = self.hdr['CDELT3']
        length = self.hdr['NAXIS3']

        # Get the exponent
        self.xexponent = get_eng_exponent(start)

        # Normalise start and increment to the exponent
        norm_factor = 10**self.xexponent
        start /= norm_factor
        increment /= norm_factor

        # Calculate end point
        end = start + length * increment

        # Create axis list
        frequency = np.linspace(start, end, length)
        return frequency

    def get_flux(self):
        """ 
        For every frequency channel, find the flux and associated uncertainty at a position
        
        Returns
        -------
        flux : list
            A list of flux values from each frequency channel
        
        f_uncert : list
            A list of flux uncertainty values for each flux measurement        
        """

        # Initialise array of fluxes and uncertainties to be returned
        flux = []
        f_uncert = []
        
        # Calculate area of the beam
        barea = _calc_beam_area(self.bvalue)
        pix2deg = self.hdr['CDELT2'] # Pixel to degree conversion factor

        # The position to find the flux at
        position = wcs2pix(self.ra, self.dec, self.hdr)
        
        # For every page of the 3D data matrix, find the flux around a point (aperture)
        for channel in self.data:

            # Setup the apertures 
            aperture = CircularAperture(position, self.aperture_radius)
            annulus = CircularAnnulus(position, r_in=2*self.aperture_radius, r_out=3*self.aperture_radius)

            # Uncertainty
            aperstats = ApertureStats(channel, annulus) 
            rms  = aperstats.std 

            # Background
            bkg = Background2D(channel, (50, 50)).background 

            # Aperture sum of the fits image minus the background
            apphot = aperture_photometry(channel - bkg, aperture)
            apsum = apphot['aperture_sum'][0]

            # Calculate corrected flux
            total_flux = apsum*(pix2deg**2)/barea

            # Save values
            flux.append(total_flux)
            f_uncert.append(rms)

        # Average 0's from values left & right
        f_uncert = average_zeroes(f_uncert) 
        
        # Normalising factor
        self.yexponent = get_eng_exponent(np.max(flux))
        norm_factor = 10**-self.yexponent

        flux = np.array(flux)*norm_factor
        f_uncert = np.array(f_uncert)*norm_factor
        return flux, f_uncert
    
    def random_analysis(self, num_points=100, radius=50, min_spread=1):
        """ 
        Iterate through n randomly generated coordinates to find the 
        signal-to-noise ratio, number of pixels, and channel peaks for each.
        
        Parameters
        ----------
        num_points : int, optional
            The number of points to find statistics for. Default = 100
        
        radius : float, optional
            The radius of the image to find statistics (in pixels) centred on
            the given ra and dec. Default = 50
        
        min_spread : float, optional
            The minimum spread of random points (in pixels). Default = 1
        
        Returns
        -------
        snrs : list
            A list of signifcant point signal-to-noise ratios
            
        pixels : list
            The list of pixels for each random position
        
        peaks : list
            The channel location of all peaks in every position
        """
        
        # Initialise return arrays
        all_snrs = []
        all_pixels = []
        all_spec_peaks = []

        # Find x,y coordinates of the target
        x, y = wcs2pix(self.ra, self.dec, self.hdr)

        # Calculate the coordinates for the number of points
        self.coordinates = spaced_circle_points(num_points, radius, [x,y], min_spread)

        # Use sslf to find lines on all points
        for i, (x, y) in enumerate(self.coordinates):
            start = time()
            self.ra, self.dec = pix2wcs(x,y, self.hdr) # convert x, y to ra,dec

            flux, uncert = self.get_flux()
            
            spec_peaks, snrs, scales = find_lines(flux) # use flux axis to find lines on
            
            pixels = calc_pixels(spec_peaks, flux)

            all_snrs.append(snrs)
            all_pixels.append(pixels)
            all_spec_peaks.append(spec_peaks)
            
            end = time()
            elapsed = end - start 
            remaining = datetime.timedelta(seconds=round(elapsed*(num_points-(i+1))))
            
            print(f'{i+1}/{len(self.coordinates)}, took {round(elapsed,2)} seconds, approx {remaining} remaining')

        return all_snrs, all_pixels, all_spec_peaks
    
    def get_exponents(self):
        """ get the x and y factors that were used to normalise the data. 
            -9 = nano, -6 = micro, -3 = milli, 3 = kilo, 6 = mega, 9 = giga,
            etc, etc, etc.
        
        Returns
        -------
        x, y : int
            The x and y normalisation factors
        """
        return self.xexponent, self.yexponent


from setuptools import setup, find_packages
import codecs
import os

VERSION = '0.0.1.3'
DESCRIPTION = 'Redshift finding algorithm'
LONG_DESCRIPTION = 'Find the redshift of high redshift radio galaxies'

# Setting up
setup(
    name="zfinder",
    version=VERSION,
    author="Daniel Lyon",
    author_email="daniellyon31@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['numpy', 'matplotlib', 'scipy', 'astropy', 'photutils', 'PyAstronomy', 'sslf'],
    keywords=['python', 'visualisation', 'redshift', 'galaxy-evolution', 'cosmology', 
              'epoch-of-reionisation', 'radio-galaxies', 'active-galactic-nuclei',
              'high-redshift', 'galaxies', 'black-holes'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)